/*
 * Copyright (c) 2016, Thibaud Ehret <ehret.thibaud@gmail.com>
 *                     Rafael Grompone for the log_gamma functions
 * All rights reserved.
 *
 * This program is free software: you can use, modify and/or
 * redistribute it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later
 * version. You should have received a copy of this license along
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file nfa.cpp
 * @brief NFA functions.
 *
 * @author Thibaud Ehret <ehret.thibaud@gmail.com>
 **/

#include "nfa.h"

#include <math.h>
#include <omp.h>
#include <iostream>
#include <stdlib.h>
#include <fstream>
#include <string>
#include <sstream>
#include <stack>
#include <algorithm>
#include <Utilities.h>
//#include <boost/math/special_functions/erf.hpp>

#define SZ_TEST2 10
#define SQRT2 1.4142135624

using namespace std;

/** Computes the natural logarithm of the absolute value of
    the gamma function of x using the Lanczos approximation.
    See http://www.rskey.org/gamma.htm

    The formula used is
    @f[
      \Gamma(x) = \frac{ \sum_{n=0}^{N} q_n x^n }{ \Pi_{n=0}^{N} (x+n) }
                  (x+5.5)^{x+0.5} e^{-(x+5.5)}
    @f]
    so
    @f[
      \log\Gamma(x) = \log\left( \sum_{n=0}^{N} q_n x^n \right)
                      + (x+0.5) \log(x+5.5) - (x+5.5) - \sum_{n=0}^{N} \log(x+n)
    @f]
    and
      q0 = 75122.6331530,
      q1 = 80916.6278952,
      q2 = 36308.2951477,
      q3 = 8687.24529705,
      q4 = 1168.92649479,
      q5 = 83.8676043424,
      q6 = 2.50662827511.
 */
static long double log_gamma_lanczos(long double x)
{
  static double q[7] = { 75122.6331530, 80916.6278952, 36308.2951477,
                         8687.24529705, 1168.92649479, 83.8676043424,
                         2.50662827511 };
  long double a = (x+0.5) * log(x+5.5) - (x+5.5);
  long double b = 0.0;
  int n;

  for(n=0;n<7;n++)
    {
      a -= log( x + (long double) n );
      b += q[n] * pow( x, (long double) n );
    }
  return a + log(b);
}

/*----------------------------------------------------------------------------*/
/** Computes the natural logarithm of the absolute value of
    the gamma function of x using Windschitl method.
    See http://www.rskey.org/gamma.htm

    The formula used is
    @f[
        \Gamma(x) = \sqrt{\frac{2\pi}{x}} \left( \frac{x}{e}
                    \sqrt{ x\sinh(1/x) + \frac{1}{810x^6} } \right)^x
    @f]
    so
    @f[
        \log\Gamma(x) = 0.5\log(2\pi) + (x-0.5)\log(x) - x
                      + 0.5x\log\left( x\sinh(1/x) + \frac{1}{810x^6} \right).
    @f]
    This formula is a good approximation when x > 15.
 */
static long double log_gamma_windschitl(long double x)
{
  return 0.918938533204673 + (x-0.5)*log(x) - x
         + 0.5*x*log( x*sinh(1/x) + 1/(810.0*pow(x,6.0)) );
}

/*----------------------------------------------------------------------------*/
/** Computes the natural logarithm of the absolute value of
    the gamma function of x. When x>15 use log_gamma_windschitl(),
    otherwise use log_gamma_lanczos().
 */
#define log_gamma(x) ((x)>15.0?log_gamma_windschitl(x):log_gamma_lanczos(x))

int computeConnectedComponent(Video<float> binary, vector<vector<unsigned> >& coords, vector<unsigned>& size)
{
	Video<int> mask(binary.sz.width, binary.sz.height, binary.sz.frames, 1, 0);

	//need a stack for the current 

	int nb = 0;
	stack<unsigned> stack;

	for(unsigned x = 0; x < binary.sz.width; ++x)
	for(unsigned y = 0; y < binary.sz.height; ++y)
	for(unsigned t = 0; t < binary.sz.frames; ++t)
	{
		if((mask(x,y,t,0) == 0) && binary(x,y,t)==1 )
		{
			nb++;
			int count = 1;
			int idx = coords.size();
			coords.push_back(vector<unsigned>());
			coords[idx].push_back(binary.sz.index(x,y,t));

			stack.push(binary.sz.index(x,y,t));
			while(!stack.empty())
			{
				unsigned index = stack.top();
				stack.pop();

				unsigned a,b,c,d;
				binary.sz.coords(index, a,b,c,d);

				if((a < binary.sz.width - 1) && (mask(a+1,b,c)==0) && binary(a+1,b,c)>0)
				{
					stack.push(binary.sz.index(a+1,b,c,0));
					coords[idx].push_back(binary.sz.index(a+1,b,c,0));
					count++;
					mask(a+1,b,c) = 1;
				}
				if((a > 0) && (mask(a-1,b,c)==0) && (binary(a-1,b,c)>0))
				{
					stack.push(binary.sz.index(a-1,b,c,0));
					coords[idx].push_back(binary.sz.index(a-1,b,c,0));
					count++;
					mask(a-1,b,c) = 1;
				}
				if((b < binary.sz.height - 1) && (mask(a,b+1,c,0)==0) && (binary(a,b+1,c)>0))
				{
					stack.push(binary.sz.index(a,b+1,c,0));
					coords[idx].push_back(binary.sz.index(a,b+1,c,0));
					count++;
					mask(a,b+1,c) = 1;
				}
				if((b > 0) && (mask(a,b-1,c,0)==0) && (binary(a,b-1,c)>0))
				{
					stack.push(binary.sz.index(a,b-1,c,0));
					coords[idx].push_back(binary.sz.index(a,b-1,c,0));
					count++;
					mask(a,b-1,c) = 1;
				}
				//if((a > 0 && b > 0) && (mask(a-1,b-1,c,0)==0) && binary(a-1,b-1,c,0)>0)
				//{
				//	stack.push(binary.sz.index(a-1,b-1,c,0));
				//	coords[idx].push_back(binary.sz.index(a-1,b-1,c,0));
				//	count++;
				//	mask(a-1,b-1,c) = 1;
				//}
				//if((a > 0 && b < binary.sz.height - 1) && (mask(a-1,b+1,c,0)==0) && binary(a-1,b+1,c,0)>0)
				//{
				//	stack.push(binary.sz.index(a-1,b+1,c,0));
				//	coords[idx].push_back(binary.sz.index(a-1,b+1,c,0));
				//	count++;
				//	mask(a-1,b+1,c) = 1;
				//}
				//if((a < binary.sz.width - 1 && b > 0) && (mask(a+1,b-1,c,0)==0) && binary(a+1,b-1,c,0)>0)
				//{
				//	stack.push(binary.sz.index(a+1,b-1,c,0));
				//	coords[idx].push_back(binary.sz.index(a+1,b-1,c,0));
				//	count++;
				//	mask(a+1,b-1,c) = 1;
				//}
				//if((a < binary.sz.width - 1 && b < binary.sz.height - 1) && (mask(a+1,b+1,c,0)==0) && binary(a+1,b+1,c,0)>0)
				//{
				//	stack.push(binary.sz.index(a+1,b+1,c,0));
				//	coords[idx].push_back(binary.sz.index(a+1,b+1,c,0));
				//	count++;
				//	mask(a+1,b+1,c) = 1;
				//}
			}
			size.push_back(count);
		}
		mask(x,y,t) = 1;
	}

	return nb;
}

float estimateDistribution(Video<float> diff, int k1, int k2, long double& proba1, long double& proba2)
{
	const float p = 0.1; // Considered as best in ipol Ponomarenko review
	const int sP = 8; // Size patches for Ponomarenko like method

	int totalPatches = (diff.sz.width - sP + 1)*(diff.sz.height - sP + 1)*diff.sz.frames;
	int quantile = totalPatches*p;
	std::vector<std::pair<float, unsigned> > sigmaP(totalPatches);

	Video<float> test(diff.sz.width, diff.sz.height, diff.sz.frames, 1, 0.f);

	// Estimate the variance for each patches 
	for(unsigned x = 0, u = 0; x < diff.sz.width - sP + 1; ++x)
	for(unsigned y = 0; y < diff.sz.height - sP + 1; ++y)
	for(unsigned t = 0; t < diff.sz.frames; ++t, ++u)
	{
		float temp = 0.f;
		for(unsigned px = 0, v = 1; px < sP; ++px)
		for(unsigned py = 0; py < sP; ++py)
		for(unsigned pc = 0; pc < diff.sz.channels; ++pc, ++v)
			temp += (diff(x+px,y+py,t,pc)*diff(x+px,y+py,t,pc) - temp) / (float)v;

		sigmaP[u] = std::make_pair(temp, test.sz.index(x,y,t));
	}

	//test.saveVideo("test.tiff", 1, 1);

	// Select the quantile patches for the estimation of the noise
	//std::partial_sort(sigmaP.begin(), sigmaP.end(), sigmaP.end()); 
	std::partial_sort(sigmaP.begin(), sigmaP.begin()+quantile, sigmaP.end(), comparaisonFirst); 
	for(unsigned k = 0; k < quantile; ++k)
	{
		unsigned x,y,t;
		test.sz.coords(sigmaP[k].second, x, y, t);
		for(unsigned px = 0; px < sP; ++px)
		for(unsigned py = 0; py < sP; ++py)
		for(unsigned pc = 0; pc < diff.sz.channels; ++pc)
			test(x+px, y+py, t) = 1;
	}
	//for(unsigned k = 0; k < totalPatches; ++k)
	//	printf("%f ", sigmaP[k]);
	//printf("\n");
	float ponsigma = sqrt(sigmaP[quantile/2].first);

	float sigma = 0.;
	for(unsigned x = 0, u = 1; x < diff.sz.width; ++x)
	for(unsigned y = 0; y < diff.sz.height; ++y)
	for(unsigned t = 0; t < diff.sz.frames; ++t)
	for(unsigned c = 0; c < diff.sz.channels; ++c, ++u)
		sigma += ((diff(x, y, t, c)*diff(x, y, t, c)) - sigma)  / (float)u;

	sigma = sqrt(sigma);
	printf("Current sigma: %f, Ponomarenko sigma: %f\n", sigma, ponsigma);
	//sigma = max(sigma, 1.f); // In case of the estimation is too small to do anything ? FIXME
	//sigma = ponsigma; // In case of the estimation is too small to do anything ? FIXME

	// Compute the two probas needed for the rest of the computations
	unsigned count1 = 0;
	unsigned count2 = 0;
	unsigned count = 0;
	for(unsigned x = 0; x < diff.sz.width; ++x)
	for(unsigned y = 0; y < diff.sz.height; ++y)
	for(unsigned t = 0; t < diff.sz.frames; ++t)
	for(unsigned c = 0; c < diff.sz.channels; ++c)
	{
		//if((sigma != ponsigma) || (test(x,y,t) > 0))
		//{	
		//	if(diff(x,y,t,c) > k1 * sigma) count1++;
		//	if(diff(x,y,t,c) > k2 * sigma) count2++;
		//	count++;
		//}
		if(diff(x,y,t,c) > k1 * sigma) count1++;
		if(diff(x,y,t,c) > k2 * sigma) count2++;
		count++;
	}
	//proba1 = (long double) count1 / (diff.sz.whcf);
	//proba2 = (long double) count2 / (diff.sz.whcf);
	printf("count %d, count1 %d, count2 %d\n", count, count1, count2);
	proba1 = (long double) count1 / count;
	proba2 = (long double) count2 / count;
	//proba1 = SQRT2 * boost::math::erfc(k1 / SQRT2);
	//proba2 = SQRT2 * boost::math::erfc(k2 / SQRT2);
	return sigma;
}

unsigned computeMaxSizeGrp(long double p)
{
	unsigned size = 0;
	long double curr = p;
	long double mini = 2.;
	while(curr < mini)
	{
		mini = curr;
		size++;
		curr = p * (size+1) + log_gamma(size+2);
	}
	return size;
}

long double computeNFAs(Video<float> original, Video<float> diff, int k1, int k2, Video<float>& visual_nfa, int& type, long double nfa_threshold, int tests)
{

	long double log_nfa_threshold = log(nfa_threshold);

	long double proba1, proba2;
	float sigma = estimateDistribution(diff, k1, k2, proba1, proba2);
	cout << "Proba1: " << proba1 << " , Proba2: " << proba2 << "\n";
	printf("%f\n", sigma);
	//int nb_tests = (original.sz.width / 20) * (original.sz.height / 20) * 4 * original.sz.frames + 1;
	int nb_tests = 2 * tests;

	/// TYPE 1
	bool type1 = false;
	long double nfa1 = 1000000000;

	// Compute connected components
	Video<float> binary(diff.sz.width, diff.sz.height, diff.sz.frames, 1, 0);

	for(unsigned x = 0; x < binary.sz.width; ++x)
	for(unsigned y = 0; y < binary.sz.height; ++y)
	for(unsigned t = 0; t < binary.sz.frames; ++t)
	for(unsigned c = 0; c < diff.sz.channels; ++c)
	{
		binary(x,y,t) += (diff(x,y,t,c) < k1 * sigma) ? 0 : 1;
	}
	binary.saveVideo("binary.tiff", 1, 1, 1);

	vector<vector<unsigned> > coordsComp;
	vector<unsigned> sizeComp;
	int nbComp = computeConnectedComponent(binary, coordsComp, sizeComp);

	long double log_tests = log(nb_tests);
	long double log_size = log(original.sz.width * original.sz.height * original.sz.frames);
	proba1 = log(diff.sz.channels * proba1);
	unsigned max_size = computeMaxSizeGrp(proba1);
	// compute the NFA for each component
	for(int comp = 0; comp < nbComp; ++comp)
	{
		sizeComp[comp] = std::min(sizeComp[comp], max_size);
		long double prob_failure = proba1 * sizeComp[comp] + log_gamma(sizeComp[comp]+1) + log_size + log_tests;
		unsigned x,y,t,c;
		binary.sz.coords(coordsComp[comp][0], x,y,t,c);
		nfa1 = min(nfa1, prob_failure);
		if(prob_failure < log_nfa_threshold)
		{
			type1 = true;
			for(unsigned u = 0; u < coordsComp[comp].size(); u++)
			{
				unsigned x,y,t,c;
				binary.sz.coords(coordsComp[comp][u], x,y,t,c);
				visual_nfa(x,y,t,0) = 255.;
			}
		}
	}

	///// TYPE 2
	bool type2 = false;
	long double nfa2 = 1000000000;

	long double opp_proba2 = log(1.-proba2);
	proba2 = log(diff.sz.channels * proba2);

	for(unsigned x = 0; x < (original.sz.width - SZ_TEST2); x+=(SZ_TEST2/2))
	for(unsigned y = 0; y < (original.sz.height - SZ_TEST2); y+=(SZ_TEST2/2))
	for(unsigned t = 0; t < original.sz.frames; ++t)
	{
		int count = 0;
		for(unsigned px = 0; px < SZ_TEST2; px++)
		for(unsigned py = 0; py < SZ_TEST2; py++)
		{
			bool activated = false;
			for(unsigned c = 0; c < original.sz.channels; ++c)
			{
				activated = (diff(x+px, y+py, t, c) < k2 * sigma) ? activated : true;
			}
			if(activated)
				count++;
		}

		//long double prob_failure = log_gamma((long double) SZ_TEST2*SZ_TEST2 + 1.0) 
		//	- log_gamma((long double) count + 1.0) 
		//	- log_gamma((long double) (SZ_TEST2*SZ_TEST2-count) + 1.0)
		//	+ count * proba2 + (SZ_TEST2*SZ_TEST2-count) * opp_proba2 + log_tests + log_size; 
		long double prob_failure = log_gamma((long double) SZ_TEST2*SZ_TEST2 + 1.0) 
			- log_gamma((long double) count + 1.0) 
			- log_gamma((long double) (SZ_TEST2*SZ_TEST2-count) + 1.0)
			+ count * proba2 + log_tests + log_size; 

		nfa2 = min(nfa2, prob_failure);
		if(prob_failure < log_nfa_threshold)
		{
			type2 = true;
			for(unsigned px = 0; px < SZ_TEST2; px++)
			for(unsigned py = 0; py < SZ_TEST2; py++)
				visual_nfa(x+px, y+py, t, 2) = 255.;
		}
	}

	for(unsigned x = 0; x < original.sz.width; ++x)
	for(unsigned y = 0; y < original.sz.height; ++y)
	for(unsigned t = 0; t < original.sz.frames; ++t)
		visual_nfa(x,y,t,1) = 255. - max(visual_nfa(x,y,t,0), visual_nfa(x,y,t,2));

	if(type1)
		if(type2)
			type = 3;
		else
			type = 1;
	else if(type2)
		type = 2;
	else
		type = 0;

	return min(nfa1, nfa2);
}
