/*
 * Original work: Copyright (c) 2013, Marc Lebrun <marc.lebrun.ik@gmail.com>
 * Modified work: Copyright (c) 2014, Pablo Arias <pariasm@gmail.com>
 * Modified work: Copyright (c) 2016, Thibaud Ehret <ehret.thibaud@gmail.com>
 * All rights reserved.
 *
 * This program is free software: you can use, modify and/or
 * redistribute it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later
 * version. You should have received a copy of this license along
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>

#include <string>
#include <sstream>
#include <sys/stat.h>

#include "Utilities/Utilities.h"
#include "NlBayes/VideoNLBayes.h"
#include "Utilities/cmd_option.h"
#include "Utilities/PatchManager/normalizedPatchManager.h"
#include "Utilities/PatchManager/videoPatchManager.h"

using namespace std;

/**
 * @file   main_extDenois.cpp
 * @brief  Main executable file for external denoising
 *
 *
 *
 * @author MARC LEBRUN  <marc.lebrun.ik@gmail.com>
 * @author PABLO ARIAS  <pariasm@gmail.com>
 * @author THIBAUD EHRET  <ehret.thibaud@gmail.com>
 **/


void rescale(Video<float>& v)
{
	for(unsigned f = 0; f < v.sz.frames; ++f)
	{
		for(unsigned c = 0; c < v.sz.channels; ++c)
		{
			// Compute the min and max
			float mm = v(0,0,f,c);
			float mM = v(0,0,f,c);
			for(unsigned x = 0; x < v.sz.width; ++x)
				for(unsigned y = 0; y < v.sz.height; ++y)
				{
					if(v(x,y,f,c) < mm)
						mm = v(x,y,f,c);
					if(v(x,y,f,c) > mM)
						mM = v(x,y,f,c);
				}
			// map the min to 0 and the max to 255.
			float coeff = 255. / (mM - mm);
			for(unsigned x = 0; x < v.sz.width; ++x)
				for(unsigned y = 0; y < v.sz.height; ++y)
				{
					v(x,y,f,c) = ((v(x,y,f,c) - mm) * coeff + .5);
				}
		}
	}	
}


int main(int argc, char **argv)
{
	clo_usage("NL-Bayes image denoising using an external database");
	clo_help(" NOTE: Input (<) and output (>) sequences are specified by their paths in printf format.\n");

	//! Paths to input/output sequences
	using std::string;
	const string  input_path = clo_option("-i"    , ""              , "< input image (multiple images can be denoised at the same time using the C patch specification such as images%03d.tiff, parameter f and l are then used for the exact specification, all images must have the same size)");
	const string  og_path = clo_option("-og"    , input_path.c_str(), "< original image, if not provided input image is used");
	//const string  inbsc_path = clo_option("-b"    , ""              , "< previous estimation of the input image");
	const string  inbsc_path = "";
	const string  data_path = clo_option("-data"    , ""            , "< database of images (multiple images can be denoised at the same time using the C patch specification such as images%03d.tiff, parameter fd and ld are then used for the exact specification, all images must have the same size)");
	const string  noisy_path = clo_option("-nisy" , "nisy_%03d.png" , "> noisy sequence");
	const string  final_path = clo_option("-deno" , "deno_%03d.png" , "> denoised sequence");
	const string  basic_path = clo_option("-bsic" , "bsic_%03d.png" , "> basic denoised sequence");
	const string    hea_path = clo_option("-heat" , "heat_%03d.tiff" , "> heatmap based on the mahalanobis weights");
	const string   diff_path = clo_option("-diff" , "diff_%03d.png" , "> difference sequence");
	const string   bias_path = clo_option("-bdeno", "bdeno_%03d.png", "> bias sequence");
	const string bbasic_path = clo_option("-bbsic", "bbsic_%03d.png", "> bias basic sequence");
	const string  bdiff_path = clo_option("-bdiff", "bdiff_%03d.png", "> bias difference sequence");

	const string   rediff_path = clo_option("-rdiff" , "rdiff_%03d.png" , "> rescaled difference sequence");

	const unsigned firstFrame = clo_option("-f", 0, "< first image");
	const unsigned lastFrame  = clo_option("-l", 0, "< last image");
	//const unsigned frameStep  = clo_option("-s", 1, "< frame step");
	const unsigned frameStep  = 1;
	const unsigned firstFrameData = clo_option("-fd", 0, "< first image from the dataset");
	const unsigned lastFrameData  = clo_option("-ld", 0, "< last image from the dataset");

	//! General parameters
	float sigma = clo_option("-sigma", 1000, "< Add noise of standard deviation sigma");
	const bool do_bias  = (bool) clo_option("-compute-bias", false, "> compute bias outputs");
	const bool verbose  = (bool) clo_option("-verbose"     , true , "> verbose output");
	const unsigned print_prms = (unsigned) clo_option("-print-prms", 0, "> prints parameters for given channels");
	const bool same  = (bool) clo_option("-same"     , true , "< if same images");

	//const unsigned tpM = clo_option("-tpM", 1, "< type mean");
	//const unsigned tpV = clo_option("-tpV", 1, "< type variance");
	//const unsigned tpdb = clo_option("-tpdb", 0, "< type variance");
	unsigned tpdb = 0;
	unsigned tpM = 0;
	unsigned tpV = 1;

	//! Video NLB parameters
	const bool flat_area1 = false;
	const bool flat_area2 = (bool) clo_option("-flat-area", false , "> use flat area trick");
	const int time_search1  = -1;
	const int time_search2  =  0;
	const int space_search1 = -1;
	const int space_search2 = clo_option("-wx",-1, "> Search window spatial radius");
	const int patch_sizex1  = -1;
	const int patch_sizex2  = clo_option("-px",-1, "> Spatial patch size");
	const int patch_sizet1  =  1;
	const int patch_sizet2  =  1;
	const int num_patches1  = -1;
	const int num_patches2  = clo_option("-np",-1, "> Number of similar patches");
	const int rank1         =  4;
	const int rank2         = clo_option("-r" , 4, "> Rank or covariance matrix");

	const int type		= clo_option("-tp", 0, "> type of search (0: local search, 1: KD-tree search, 2: VP-tree search, 3: VPMP, 4: VPLR)");
	const int search	= clo_option("-search", 0, "> type of search (0: local search, 1: KD-tree search, 2: VP-tree search, 3: VPMP, 4: VPLR)");
	//const int manager	= clo_option("-mngr", 0, "> type of patch manager ");
	const int manager	= 0;
	//const int reranking	= clo_option("-rr", 0, "> reranking of the elements if possible");
	const int reranking	= 0;
	const int nbTrees	= clo_option("-nbT", 1, "> number of trees");
	const float flatThreshold = clo_option("-flat", 0, "> threshold propbabilities for the detection of flat patches (0 < threshold in peeercent < 100)") / 100.f;
	//const float flatThreshold = 0.9;
	const unsigned darkRegion = clo_option("-DR", 0, "< Threshold at which the region are consider dark and need an other algorithm");


	//! Only print parameters
	if (print_prms)
	{
		VideoSize tmp;
		tmp.channels = print_prms;

		//! Compute denoising default parameters
		VideoNLB::nlbParams prms1, prms2;
		VideoNLB::initializeNlbParameters(prms1, 1, sigma, tmp, flat_area1, verbose, time_search1, time_search1, patch_sizet1);
		VideoNLB::initializeNlbParameters(prms2, 2, sigma, tmp, flat_area2, verbose, time_search2, time_search2, patch_sizet2);

		//! Override with command line parameters
		if (space_search1 >= 0) VideoNLB::setSizeSearchWindow(prms1, (unsigned)space_search1);
		if (space_search2 >= 0) VideoNLB::setSizeSearchWindow(prms2, (unsigned)space_search2);
		if (patch_sizex1  >= 0) VideoNLB::setSizePatch(prms1, tmp, (unsigned)patch_sizex1);;
		if (patch_sizex2  >= 0) VideoNLB::setSizePatch(prms2, tmp, (unsigned)patch_sizex2);;
		if (num_patches1  >= 0) VideoNLB::setNSimilarPatches(prms1, (unsigned)num_patches1);
		if (num_patches2  >= 0) VideoNLB::setNSimilarPatches(prms2, (unsigned)num_patches2);

		prms1.rank = rank1;
		prms2.rank = rank2;

		VideoNLB::printNlbParameters(prms1);
		VideoNLB::printNlbParameters(prms2);

		return EXIT_FAILURE;
	}


	//! Check inputs
	if (input_path == "")
	{
		fprintf(stderr, "%s: no input sequence.\nTry `%s --help' for more information.\n",
				argv[0], argv[0]);
		return EXIT_FAILURE;
	}

	if (patch_sizex1 == 0 && inbsc_path == "")
	{
		fprintf(stderr, "%s: if px1 = 0, a basic sequence path must be given.\nTry `%s --help' for more information.\n",
				argv[0], argv[0]);
		return EXIT_FAILURE;
	}

	if (patch_sizex1 > 0 && inbsc_path != "")
		fprintf(stderr, "\x1b[33;1mWarning:\x1b[0m a basic sequence path ignored since px1 > 0.\n");

	printf("same %d\n", same);

	//! Declarations
	Video<float> original, noisy, basic, final, diff, heatmap;

	//! Load input videos
	                       original.loadVideo(input_path, firstFrame, lastFrame, frameStep);
	if (patch_sizex1 == 0) basic   .loadVideo(inbsc_path, firstFrame, lastFrame, frameStep);
	Video<float> data;
	data.loadVideo(data_path, firstFrameData, lastFrameData, 1);

	noisy = original;

	//! Add noise
	//if (sigma)
	//{
	//	VideoUtils::addNoise(original, noisy, sigma, verbose);

	//	//! Save noisy video
	//	if (verbose) printf("Saving noisy video\n");
	//	noisy.saveVideo(noisy_path, firstFrame, frameStep);
	//}
	//printf("Remove this part regarding the noise");
	//// If the initial noise is not negligeable compared to the one we're adding
	//if(sigma < 5)
	//	sigma = std::sqrt(1+sigma);

	//! Denoising
	if (verbose) printf("Running Video NL-Bayes on the noisy video\n");

	//! Compute denoising default parameters
	VideoNLB::nlbParams prms1, prms2;
	VideoNLB::initializeNlbParameters(prms1, 1, sigma, noisy.sz, flat_area1, verbose, time_search1, time_search1, patch_sizet1);
	VideoNLB::initializeNlbParameters(prms2, 2, sigma, noisy.sz, flat_area2, verbose, time_search2, time_search2, patch_sizet2);

	//! Override with command line parameters
	if (space_search1 >= 0) VideoNLB::setSizeSearchWindow(prms1, (unsigned)space_search1);
	if (space_search2 >= 0) VideoNLB::setSizeSearchWindow(prms2, (unsigned)space_search2);
	if (patch_sizex1  >= 0) VideoNLB::setSizePatch(prms1, noisy.sz, (unsigned)patch_sizex1);;
	if (patch_sizex2  >= 0) VideoNLB::setSizePatch(prms2, noisy.sz, (unsigned)patch_sizex2);;
	if (num_patches1  >= 0) VideoNLB::setNSimilarPatches(prms1, (unsigned)num_patches1);
	if (num_patches2  >= 0) VideoNLB::setNSimilarPatches(prms2, (unsigned)num_patches2);

	prms1.rank = rank1;
	prms2.rank = rank2;
	prms1.type    = type;
	prms2.type    = type;
	prms1.manager = manager;
	prms2.manager = manager;
	prms1.nbTrees = nbTrees;
	prms2.nbTrees = nbTrees;
	prms1.reranking = reranking;
	prms2.reranking = reranking;
	prms1.flat = flatThreshold;
	prms2.flat = flatThreshold;
	prms1.darkRegion = darkRegion;
	prms2.darkRegion = darkRegion;
	prms1.same = same;
	prms2.same = same;
	prms1.search = search;
	prms2.search = search;

	//! Percentage or processed groups of patches over total number of pixels
	std::vector<float> groupsRatio;

	PatchManager* dbpm1;
	if(tpdb == 0)
		dbpm1 = new DatabasePatchManager(data, patch_sizex2, patch_sizet2, data.sz.channels);
	else
		dbpm1 = new DatabaseNormalizedPatchManager(data, patch_sizex2, patch_sizet2, data.sz.channels, tpM, tpV);
	PatchManager* pm;
	if(tpdb == 0)
		pm = new VideoPatchManager(data, patch_sizex2, patch_sizet2, data.sz.channels);
	else
		pm = new NormalizedPatchManager(data, patch_sizex2, patch_sizet2, data.sz.channels, tpM, tpV);

	std::vector<PartitionTree*> trees; 
	//if(prms2.type > 0)
	{
		PrmTree paramTree;
		paramTree.epsilon = 0.; 
		paramTree.prms = prms2;
		paramTree.nbVP = 1;
		paramTree.subS = 1000;
		paramTree.rand = 5;
#ifdef _OPENMP
#pragma omp parallel for schedule(dynamic) \
		shared(trees) 
#endif
		for(int i = 0; i < prms2.nbTrees; ++i)
		{
			char filename[128];
			sprintf(filename, "vp%d.save", i);
			//if(prms2.type > 1)
				trees.push_back(new VPTree(*pm, paramTree, filename));	
			//else
			//{
			//	trees.push_back(new KDTree(pm, paramTree, filename));	
			//}
		}
		//for(int i = 0; i < prms2.nbTrees; ++i)
		//{
		//	char filename[128];
		//	sprintf(filename, "vp%d.save", i);
		//	struct stat buffer;

		//	if(stat(filename, &buffer) != 0)
		//		trees[i]->exportTree(filename);
		//}
	}
	ForestManager dbfm1(*pm, trees, prms2.nSimilarPatches, prms2);
	printf("Denoising\n");

	//! Run denoising algorithm
	groupsRatio = VideoNLB::runExtNlBayes(noisy, basic, final, heatmap, *dbpm1, *pm, dbfm1, prms2);

	original.loadVideo(og_path, firstFrame, lastFrame, frameStep);

	//! Compute PSNR and RMSE
	float final_psnr = -1, final_rmse = -1, basic_psnr = -1, basic_rmse = -1;
	if (prms2.sizePatch) VideoUtils::computePSNR(original, final, final_psnr, final_rmse);
	                     VideoUtils::computePSNR(original, basic, basic_psnr, basic_rmse);

	if (verbose)
	{
	    printf("basic PSNR =\t%f\tRMSE =\t%f\n", basic_psnr, basic_rmse);
	    printf("final PSNR =\t%f\tRMSE =\t%f\n", final_psnr, final_rmse);
	}

	//! Write measures
	//writingMeasures("measures.txt", sigma, basic_psnr, basic_rmse, groupsRatio[0], groupsRatio[1], groupsRatio[2], true  , "_basic");
	writingMeasures("measures.txt", sigma, final_psnr, final_rmse, groupsRatio[3], groupsRatio[4], groupsRatio[5], false , "_final");

	//! Compute Difference
	if (prms2.sizePatch) VideoUtils::computeDiff(original, final, diff, sigma);
	
	// Rescale the difference for visualisation purposes
	Video<float> re_diff = diff;
	rescale(re_diff);

	for(int i = 0; i < prms2.nbTrees; ++i)
		delete trees[i];
	delete dbpm1;
	delete pm;

	//! Save output sequences
	if (verbose) printf("Saving output sequences\n");

	Video<float> re_heat = heatmap;
	//rescale(re_heat);

	if (prms2.sizePatch) final  .saveVideo(final_path,  firstFrame, frameStep);
	if (prms2.sizePatch) diff   .saveVideo(diff_path,   firstFrame, frameStep);
	if (prms2.sizePatch) re_diff.saveVideo(rediff_path, firstFrame, frameStep);
	                     basic  .saveVideo(basic_path,  firstFrame, frameStep);
	if((prms2.type == 4) || (prms2.type == 5))
	                     re_heat.saveVideo(rediff_path,    firstFrame, frameStep);

#if 0
	//! Computing bias sequence
	Video<float> bias, bias_basic, bias_diff;
	if (do_bias) {
		if (verbose) cout << "Applying NL-Bayes to the original image :" << endl;

		runNlBayes(im, imBasicBias, imBias, imSize, flat_area1, flat_area2, sigma, verbose);

		if (verbose) cout << endl;

		float psnrBias, psnrBiasBasic, rmseBias, rmseBiasBasic;
		computePsnr(im, imBasicBias, psnrBiasBasic, rmseBiasBasic, "imBiasBasic", verbose);
		computePsnr(im, imBias     , psnrBias     , rmseBias     , "imBiasFinal", verbose);

		//! writing measures
		writingMeasures("measures.txt", sigma, psnrBiasBasic, rmseBiasBasic, false, "_bias_basic");
		writingMeasures("measures.txt", sigma, psnrBias     , rmseBias     , false, "_bias      ");

		
		computeDiff(im, imBias, imDiffBias, sigma, 0.f, 255.f, verbose);

		saveImage(argv[7], imBias     , imSize, 0.f, 255.f);
		saveImage(argv[8], imBasicBias, imSize, 0.f, 255.f);
		saveImage(argv[9], imDiffBias , imSize, 0.f, 255.f);
	}
#endif

	if (verbose) printf("Done\n");
	return EXIT_SUCCESS;
}
